from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

from dipy.core.gradients import gradient_table
import dipy.reconst.dti as dti
import dipy.reconst.fwdti as fwdti
import nibabel as nib
from dipy import __version__ as dipy_version

IFLOGGER=nlogging.getLogger('nipype.interface')

def freewater_proc(labels_dict,input_dir):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    output_dir = cwd
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')
    if not session_label:
        search_session_label="*"
    else:
        search_session_label=session_label

    input_dir=substitute_labels(input_dir,labels_dict)
    dwi = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-preproc_dwi.nii.gz'))
    bval = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-preproc_dwi.bval'))
    bvec = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-preproc_dwi.bvec'))
    mask = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-brain_mask.nii.gz'))

    freewater_dir = os.path.join(cwd,'freewater')
    if not os.path.isdir(freewater_dir):
        os.makedirs(freewater_dir,exist_ok=True)

    freeFA_dipy = os.path.join(freewater_dir,f"sub-{participant_label}_ses-{session_label}_space-T1w_desc-preproc_rec-dipy_desc-freewater_desc-fa.nii.gz")
    freeFA_corrected_dipy = os.path.join(freewater_dir,f"sub-{participant_label}_ses-{session_label}_space-T1w_desc-preproc_rec-dipy_desc-freewater_desc-facorrected.nii.gz")
    freeMD_dipy = os.path.join(freewater_dir,f"sub-{participant_label}_ses-{session_label}_space-T1w_desc-preproc_rec-dipy_desc-freewater_desc-md.nii.gz")
    freeFrac_dipy = os.path.join(freewater_dir,f"sub-{participant_label}_ses-{session_label}_space-T1w_desc-preproc_rec-dipy_desc-freewater_desc-fraction.nii.gz")

    dti_dir = os.path.join(cwd,'dti')
    if not os.path.isdir(dti_dir):
        os.makedirs(dti_dir,exist_ok=True)

    dtiFA_dipy = os.path.join(dti_dir,f"sub-{participant_label}_ses-{session_label}_space-T1w_desc-preproc_rec-dipy_desc-fa.nii.gz")
    dtiFA_corrected_dipy = os.path.join(dti_dir,f"sub-{participant_label}_ses-{session_label}_space-T1w_desc-preproc_rec-dipy_desc-facorrected.nii.gz")
    dtiMD_dipy = os.path.join(dti_dir,f"sub-{participant_label}_ses-{session_label}_space-T1w_desc-preproc_rec-dipy_desc-md.nii.gz")

    IFLOGGER.info("References:\n\nOfer Pasternak, Nir Sochen, Yaniv Gur, Nathan Intrator, and Yaniv Assaf. Free water elimination and mapping from diffusion MRI. Magnetic Resonance in Medicine, 62(3):717–730, 2009. URL: 10.1002/mrm.22055, doi:https://doi.org/10.1002/mrm.22055\n\nAndrew R. Hoy, Cheng Guan Koay, Steven R. Kecskemeti, and Andrew L. Alexander. Optimization of a free water elimination two-compartment model for diffusion tensor imaging. NeuroImage, 103:323–333, 2014. URL: 10.1016/j.neuroimage.2014.09.053, doi:https://doi.org/10.1016/j.neuroimage.2014.09.053.\n\nRafael Neto Henriques, Ariel Rokem, Eleftherios Garyfallidis, Samuel St-Jean, Eric Thomas Peterson, and Marta Morgado Correia. [Re] Optimization of a free water elimination two-compartment model for diffusion tensor imaging. bioRxiv, 2017. doi:https://doi.org/10.1101/108795.\n")

    IFLOGGER.info("See: https://docs.dipy.org/stable/examples_built/reconstruction/reconst_fwdti.html#sphx-glr-download-examples-built-reconstruction-reconst-fwdti-py")

    dwiimg = nib.load(dwi)
    maskimg = nib.load(mask)
    gtab = gradient_table(bvals=bval,bvecs=bvec)
    dwidata = np.asarray(dwiimg.dataobj)
    maskdata = maskimg.get_fdata()

    IFLOGGER.info(f"Calculating FreeWater Tensor Model using Dipy {dipy_version}")
  
    fwdtimodel = fwdti.FreeWaterTensorModel(gtab)
    fwdtifit = fwdtimodel.fit(dwidata, mask=maskdata)
    freeFA = fwdtifit.fa
    freeMD = fwdtifit.md
    freeFrac = fwdtifit.f

    FRAC_THRESHOLD = getParams(labels_dict,'FRAC_THRESHOLD')
    if not FRAC_THRESHOLD:
        FRAC_THRESHOLD=0.7
    freeFA_corrected=freeFA.copy()
    freeFA_corrected[freeFrac > FRAC_THRESHOLD]=0

    freeFAimg = nib.Nifti1Image(freeFA, dwiimg.affine, dwiimg.header)
    nib.save(freeFAimg,freeFA_dipy) 

    freeMDimg = nib.Nifti1Image(freeMD, dwiimg.affine, dwiimg.header)
    nib.save(freeMDimg,freeMD_dipy) 

    freeFracimg = nib.Nifti1Image(freeFrac, dwiimg.affine, dwiimg.header)
    nib.save(freeFracimg,freeFrac_dipy) 

    freeFAcorrectedimg = nib.Nifti1Image(freeFA_corrected, dwiimg.affine, dwiimg.header)
    nib.save(freeFAcorrectedimg,freeFA_corrected_dipy) 

    dtimodel = dti.TensorModel(gtab)
    dtifit = dtimodel.fit(dwidata, mask=maskdata)
    dtiFA = dtifit.fa
    dtiMD = dtifit.md
    dtiFA_corrected=dtiFA.copy()
    dtiFA_corrected[freeFrac > FRAC_THRESHOLD]=0

    dtiFAimg = nib.Nifti1Image(dtiFA, dwiimg.affine, dwiimg.header)
    nib.save(dtiFAimg,dtiFA_dipy) 

    dtiFAcorrectedimg = nib.Nifti1Image(dtiFA_corrected, dwiimg.affine, dwiimg.header)
    nib.save(dtiFAcorrectedimg,dtiFA_corrected_dipy) 

    dtiMDimg = nib.Nifti1Image(dtiMD, dwiimg.affine, dwiimg.header)
    nib.save(dtiMDimg,dtiMD_dipy) 

    out_files=[]
    out_files.insert(0,freeFA_dipy)
    out_files.insert(1,freeMD_dipy)
    out_files.insert(2,dtiFA_dipy)
    out_files.insert(3,dtiMD_dipy)


    return {
        "freefa":freeFA_dipy,
        "freemd":freeMD_dipy,
        "fraction":freeFrac_dipy,
        "dtifa":dtiFA_dipy,
        "dtimd":dtiMD_dipy,
        "output_dir":output_dir,
        "out_files":out_files
    }



class freewaterInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_dir = traits.String("",desc="QSIPREP Output Directory", usedefault=True)

class freewaterOutputSpec(TraitedSpec):
    freefa = File(desc='FreeFA')
    freemd = File(desc='FreeMD')
    fraction = File(desc='FreewaterFraction')
    dtifa = File(desc='dtiFA')
    dtimd = File(desc='dtiMD')
    output_dir = traits.String(desc="Freewater output directory")
    out_files = traits.List(desc='list of files')
    
class freewater_pan(BaseInterface):
    input_spec = freewaterInputSpec
    output_spec = freewaterOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = freewater_proc(
            self.inputs.labels_dict,
            self.inputs.input_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="freewater_node",input_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(freewater_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if input_dir is None:
        input_dir = ""
        
    pan_node.inputs.input_dir =  input_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
import pandas as pd
import json
import datetime
from pathlib import Path
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

newnet7 = {
    "1" : {
        "netname": ["Vis"],
        "exclusions" : ["Cerebellum"],
        "newname" : "Cortical.Vis"
    },
    "2" : {
        "netname": ["Def"],
        "exclusions" : ["Cerebellum"],
        "newname" : "Cortical.Def"
    },
    "3" : {
        "netname": ["SomMot"],
        "exclusions" : ["Cerebellum"],
        "newname" : "Cortical.SomMot"
    },
    "4" : {
        "netname": ["Cont"],
        "exclusions" : ["Cerebellum"],
        "newname" : "Cortical.Cont"
    },
    "5" : {
        "netname": ["DorsAttn"],
        "exclusions" : ["Cerebellum"],
        "newname" : "Cortical.DorsAttn"
    },
    "6" : {
        "netname": ["Limbic"],
        "exclusions" : ["Cerebellum"],
        "newname" : "Cortical.Limbic"
    },
    "7" : {
        "netname": ["SalVentAttn"],
        "exclusions" : ["Cerebellum"],
        "newname" : "Cortical.SalVentAttn"
    },
    "8" : {
        "netname": ["Vis","Cerebellum"],
        "exclusions" : [],
        "newname" : "Cereb.Vis"
    },
    "9" : {
        "netname": ["Def","Cerebellum"],
        "exclusions" : [],
        "newname" : "Cereb.Def"
    },
    "10" : {
        "netname": ["SomMot","Cerebellum"],
        "exclusions" : [],
        "newname" : "Cereb.SomMot"
    },
    "11" : {
        "netname": ["Cont","Cerebellum"],
        "exclusions" : [],
        "newname" : "Cereb.Cont"
    },
    "12" : {
        "netname": ["DorsAttn","Cerebellum"],
        "exclusions" : [],
        "newname" : "Cereb.DorsAttn"
    },
    "13" : {
        "netname": ["Limbic","Cerebellum"],
        "exclusions" : [],
        "newname" : "Cereb.Limbic"
    },
    "14" : {
        "netname": ["SalVentAttn","Cerebellum"],
        "exclusions" : [],
        "newname" : "Cereb.SalVentAttn"
    }
}
net17= ["VisPeri","VisCent","DefaultA","DefaultB","DefaultC","TempPar","SomMotA","SomMotB","ContA","ContB","ContC","DorsAttnA","DorsAttnB","LimbicA","LimbicB","SalVentAttnA","SalVentAttnB"]
net7 = ["Vis","Def","SomMot","Cont","DorsAttn","Limbic","SalVentAttn"]

def addExtraCols(df,labels_dict,subject,session):
    df = processExtraColumns(df, labels_dict)
    if session: 
        df.insert(0,"session_id",[session for x in range(len(df))])
    if subject:
        df.insert(0,"subject_id",[subject for x in range(len(df))])
    return df

def calculate_custom_interconnectivity(corrdf,net=newnet7,nodecol="Node"):
    inter_df=pd.DataFrame()
    table_vals=[]
    table_cols=[]

    nodeidxs = net.keys()
    nodepairs = intranode_pairings(nodeidxs)
   
    for pair in nodepairs:
        idx1=pair[0]
        idx2=pair[1]
        netname1 = net[idx1]["netname"]
        netname2 = net[idx2]["netname"]
        new_netname1 = net[idx1]["newname"]
        new_netname2 = net[idx2]["newname"]
        exclusions1 = net[idx1]["exclusions"]
        exclusions2 = net[idx2]["exclusions"]

        netnodes1 = [x for x in corrdf.columns if all(y.lower() in x.lower() for y in netname1) and not any(y.lower() in x.lower() for y in exclusions1) ]
        netnodes2 = [x for x in corrdf.columns if all(y.lower() in x.lower() for y in netname2) and not any(y.lower() in x.lower() for y in exclusions2) ]
        
        if len(netnodes1) > 0 and len(netnodes2) > 0:
            intercorr = calc_inter(corrdf,netnodes1,netnodes2)
            table_cols.append(f"{new_netname1}_{new_netname2}")
            table_vals.append(intercorr)
        else:
            IFLOGGER.warn(f"Cannot calculate inter-regional connectivity with missing nodes - netnode1:{netnodes1} netnode2:{netnodes2}")
            table_cols.append(f"{new_netname1}_{new_netname2}")
            table_vals.append(np.nan)
    if table_cols and table_vals:
        inter_df = pd.DataFrame([table_vals])
        inter_df.columns = table_cols
    return inter_df

def calculate_yeobuckner_interconnectivity(corrdf,net=net7,nodecol="Node",prefix="intra",exception=None,addition=None):
    inter_df=pd.DataFrame()
    table_vals=[]
    table_cols=[]

    nodepairs = intranode_pairings(net7)
   
    for pair in nodepairs:
        netname1=pair[0]
        netname2=pair[1]
        if exception:
            netnodes1 = [ x for x in corrdf.columns if netname1.lower() in x.lower() and exception.lower() not in x.lower()]
            netnodes2 = [ x for x in corrdf.columns if netname2.lower() in x.lower() and exception.lower() not in x.lower()]

        elif addition:
            netnodes1 = [ x for x in corrdf.columns if netname1.lower() in x.lower() and addition.lower() in x.lower()]
            netnodes2 = [ x for x in corrdf.columns if netname2.lower() in x.lower() and addition.lower() in x.lower()]
        else:
            netnodes1 = [ x for x in corrdf.columns if netname1.lower() in x.lower()]
            netnodes2 = [ x for x in corrdf.columns if netname2.lower() in x.lower()]
        if len(netnodes1) > 0 and len(netnodes2) > 0:
            intercorr = calc_inter(corrdf,netnodes1,netnodes2)
            table_cols.append(f"{prefix}.{netname1}_{netname2}")
            table_vals.append(intercorr)
        else:
            IFLOGGER.warn(f"Cannot calculate inter-regional connectivity with missing nodes - netnode1:{netnodes1} netnode2:{netnodes2}")
            table_cols.append(f"{prefix}.{netname1}_{netname2}")
            table_vals.append(np.nan)
    if table_cols and table_vals:
        inter_df = pd.DataFrame([table_vals])
        inter_df.columns = table_cols
    return inter_df

def calculate_yeobuckner_intraconnectivity(corrdf,net=net7,nodecol="Node",prefix="intra",exception=None,addition=None):
    intra_df=pd.DataFrame()
    table_vals=[]
    table_cols=[]
    for netname in net:
        if exception:
            netnodes = [ x for x in corrdf.columns if netname.lower() in x.lower() and exception.lower() not in x.lower()]
        elif addition:
            netnodes = [ x for x in corrdf.columns if netname.lower() in x.lower() and addition.lower() in x.lower()]
        else:
            netnodes = [ x for x in corrdf.columns if netname.lower() in x.lower()]
        if len(netnodes) > 1:
            intracorr = calc_intra(corrdf,netnodes)
            table_cols.append(f"{prefix}_{netname}")
            table_vals.append(intracorr)
        else:
            IFLOGGER.warn(f"Cannot calculate intra-regional connectivity with 1 node - {netnodes}")
            table_cols.append(f"{prefix}_{netname}")
            table_vals.append(np.nan)
    if table_cols and table_vals:
        intra_df = pd.DataFrame([table_vals])
        intra_df.columns = table_cols
    return intra_df

def postxcpd_proc(labels_dict,input_dir):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    output_dir=cwd
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    subject = f"sub-{participant_label}"
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')
    input_dir=substitute_labels(input_dir,labels_dict)

    if not session_label:
        session=None
        postxcpd_outputdir = os.path.join(cwd,f"{subject}_postxcpd_outdir")
    else:
        session = f"ses-{session_label}"
        postxcpd_outputdir = os.path.join(cwd,f"{subject}_{session}_postxcpd_outdir")

    if not os.path.isdir(postxcpd_outputdir):
        os.makedirs(postxcpd_outputdir,exist_ok=True)

    try:
        atlases=[x for x in os.listdir(os.path.join(input_dir,"atlases")) if x.startswith("atlas-" )]
    except Exception as exc:
        atlases=[]

    for atlas in atlases:
        atlasstub = atlas.split("-")[1]

        replacecol={}
        atlas_config=getParams(labels_dict,atlasstub)
        if atlas_config:
            if "replacecol" in atlas_config.keys():
                replacecol = atlas_config["replacecol"]
        else:
            continue

        rehopath = os.path.join(input_dir,subject,session,"func",f"{subject}_{session}_*{atlasstub}*reho_bold.tsv")
        rehofile=glob.glob(rehopath)
        if rehofile:
            rehofile = rehofile[0]
            rehodf=pd.read_table(rehofile,sep="\t")
            rehodf = addExtraCols(rehodf,labels_dict,subject,session)
            reho_out = newfile(outputdir=postxcpd_outputdir,assocfile=rehofile,extension="csv")
            rehodf.to_csv(reho_out ,sep=",",header=True, index=False)

        alffpath = os.path.join(input_dir,subject,session,"func",f"{subject}_{session}_*{atlasstub}*alff_bold.tsv")
        alfffile=glob.glob(alffpath)
        if alfffile:
            alfffile = alfffile[0]
            alffdf=pd.read_table(alfffile,sep="\t")
            alffdf = addExtraCols(alffdf,labels_dict,subject,session)
            alff_out = newfile(outputdir=postxcpd_outputdir,assocfile=alfffile,extension="csv")
            alffdf.to_csv(alff_out ,sep=",",header=True, index=False)

        coverpath = os.path.join(input_dir,subject,session,"func",f"{subject}_{session}_*{atlasstub}*coverage_bold.tsv")
        coverfile=glob.glob(coverpath)
        if coverfile:
            coverfile = coverfile[0]
            coverdf=pd.read_table(coverfile,sep="\t")
            coverflatdf=transpose(coverdf)
            coverflatdf = addExtraCols(coverflatdf,labels_dict,subject,session)
            flatcover = newfile(outputdir=postxcpd_outputdir,assocfile=coverfile,extension="csv")
            coverflatdf.to_csv(flatcover,sep=",",header=True, index=False)

        corrpath = os.path.join(input_dir,subject,session,"func",f"{subject}_{session}_*{atlasstub}*relmat.tsv")
        corrfile=glob.glob(corrpath)
        if corrfile:
            corrfile = corrfile[0]
            corrdf=pd.read_table(corrfile,sep="\t")
            flatdf=flatten(corrdf,replace=replacecol)
            flatdf = addExtraCols(flatdf,labels_dict,subject,session)
            flatcorr = newfile(outputdir=postxcpd_outputdir,assocfile=corrfile,suffix="desc-flat",extension="csv")
            flatdf.to_csv(flatcorr,sep=",",header=True, index=False)
        
        # if this is a yeobuckner 131 atlas - we will calculate the intra-regional connectivity based on net7 and net17
        if "yeobuckner131" in atlasstub:
            intra_wholebrain_df = calculate_yeobuckner_intraconnectivity(corrdf,net=net7,nodecol="Node",prefix="wholeintra")
            intra_wholebrain_df = addExtraCols(intra_wholebrain_df,labels_dict,subject,session)

            intra_cortical_df = calculate_yeobuckner_intraconnectivity(corrdf,net=net7,nodecol="Node",prefix="corticalintra",exception="Cerebellum")
            yeo131_net7_intradf = pd.concat([intra_wholebrain_df,intra_cortical_df],axis=1)

            intra_cerebellum_df = calculate_yeobuckner_intraconnectivity(corrdf,net=net7,nodecol="Node",prefix="cerebintra",addition="Cerebellum")
            yeo131_net7_intradf = pd.concat([yeo131_net7_intradf,intra_cerebellum_df],axis=1)

            yeo131_net7= os.path.join(postxcpd_outputdir,f"{subject}_{session}_{atlasstub}_Networks7_intraconnectivity.csv")
            yeo131_net7_intradf.to_csv(yeo131_net7,sep=",",header=True, index=False)

            intra_wholebrain_df = calculate_yeobuckner_intraconnectivity(corrdf,net=net17,nodecol="Node",prefix="wholeintra")
            intra_wholebrain_df = addExtraCols(intra_wholebrain_df,labels_dict,subject,session)

            intra_cortical_df = calculate_yeobuckner_intraconnectivity(corrdf,net=net17,nodecol="Node",prefix="corticalintra",exception="Cerebellum")
            yeo131_net17_intradf = pd.concat([intra_wholebrain_df,intra_cortical_df],axis=1)

            yeo131_net17= os.path.join(postxcpd_outputdir,f"{subject}_{session}_{atlasstub}_Networks17_intraconnectivity.csv")
            yeo131_net17_intradf.to_csv(yeo131_net17,sep=",",header=True, index=False)

            # we also calculate the inter-regional connectivity
            inter_wholebrain_df = calculate_yeobuckner_interconnectivity(corrdf,net=net7,nodecol="Node",prefix="wholeinter")
            inter_wholebrain_df = addExtraCols(inter_wholebrain_df,labels_dict,subject,session)

            #inter_cortical_df = calculate_yeobuckner_interconnectivity(corrdf,net=net7,nodecol="Node",prefix="corticalinter",exception="Cerebellum")
            #yeo131_net7_interdf = pd.concat([inter_wholebrain_df,inter_cortical_df],axis=1)

            #inter_cerebellum_df = calculate_yeobuckner_interconnectivity(corrdf,net=net7,nodecol="Node",prefix="cerebinter",addition="Cerebellum")
            #yeo131_net7_interdf = pd.concat([yeo131_net7_interdf,inter_cerebellum_df],axis=1)

            yeo131_net7_inter= os.path.join(postxcpd_outputdir,f"{subject}_{session}_{atlasstub}_Networks7_wholebrain_interconnectivity.csv")
            inter_wholebrain_df.to_csv(yeo131_net7_inter,sep=",",header=True, index=False)

            inter_newwholebrain_df = calculate_custom_interconnectivity(corrdf,net=newnet7,nodecol="Node")
            inter_newwholebrain_df  = addExtraCols(inter_newwholebrain_df ,labels_dict,subject,session)
            inter_newwholebrain= os.path.join(postxcpd_outputdir,f"{subject}_{session}_{atlasstub}_Networks7_corticocerebellar_interconnectivity.csv")
            inter_newwholebrain_df.to_csv(inter_newwholebrain,sep=",",header=True, index=False)


        # if this is a yeobuckner 131 atlas - we will calculate the intra-regional connectivity based on net7 and net17
        if "yeobuckner58" in atlasstub:
            intra_wholebrain_df = calculate_yeobuckner_intraconnectivity(corrdf,net=net7,nodecol="Node",prefix="wholeintra")
            intra_wholebrain_df = addExtraCols(intra_wholebrain_df,labels_dict,subject,session)

            intra_cortical_df = calculate_yeobuckner_intraconnectivity(corrdf,net=net7,nodecol="Node",prefix="corticalintra",exception="Cerebellum")
            yeo58_net7_intradf = pd.concat([intra_wholebrain_df,intra_cortical_df],axis=1)

            yeo58_net7= os.path.join(postxcpd_outputdir,f"{subject}_{session}_{atlasstub}_Networks7_intraconnectivity.csv")
            yeo58_net7_intradf.to_csv(yeo58_net7,sep=",",header=True, index=False)

        atlaspath = os.path.join(input_dir,"atlases",f"{atlas}",f"{atlas}*.nii.gz")
        atlasfile=glob.glob(atlaspath)
        labelpath = os.path.join(input_dir,"atlases",f"{atlas}",f"{atlas}*.tsv")
        labelfile=glob.glob(labelpath)
        if atlasfile and labelfile:
            atlasfile=atlasfile[0]
            labelfile=labelfile[0]
            roisizes_df = roisizes(atlasfile,labelfile)
            roisizes_file = os.path.join(postxcpd_outputdir,f"{subject}_{session}_{atlasstub}_roisizes.csv")
            roisizes_df.to_csv(roisizes_file,sep=",",header=True, index=False)
        
    out_files=[]

    return {
        "output_dir":output_dir,
        "out_files":out_files
    }


class postxcpdInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_dir = traits.String("",desc="XCPD Output Directory", usedefault=True)


class postxcpdOutputSpec(TraitedSpec):
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class postxcpd_pan(BaseInterface):
    input_spec = postxcpdInputSpec
    output_spec = postxcpdOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = postxcpd_proc(
            self.inputs.labels_dict,
            self.inputs.input_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="postxcpd_node",input_dir=""):
    # Create Node
    pan_node = Node(postxcpd_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    if input_dir is None:
        input_dir = ""       
    pan_node.inputs.input_dir =  input_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from shutil import copytree
from nipype import logging as nlogging
from bids import BIDSLayout

IFLOGGER=nlogging.getLogger('nipype.interface')

IS_PRESENT="^^^"
IGNORE="###"

def mriqcgroup_proc(labels_dict,bids_dir=""):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    mriqcgroup_outdir=os.path.join(cwd,"mriqcgroupout")
    os.makedirs(mriqcgroup_outdir,exist_ok=True)

    command_base, container = getContainer(labels_dict,nodename="mriqc", SPECIFIC="MRIQC_CONTAINER",LOGGER=IFLOGGER)
    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=translate_binding(command_base,TEMPLATEFLOW_HOME)

    IFLOGGER.info("Checking the mriqc version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    
    participants_label = getParams(labels_dict,'GROUP_PARTICIPANTS_LABEL')
    participants_project = getParams(labels_dict,'GROUP_PARTICIPANTS_XNAT_PROJECT')
    participants_session = getParams(labels_dict,'GROUP_SESSION_LABEL')
    subject_list =[]
    if participants_label:
        for part_vals in zip(participants_project,participants_label,participants_session):
            project= part_vals[0]
            subject=part_vals[1]
            session = part_vals[2]
            subject_list.append(f"{project}^{subject}^{session}")

    mriqc_sourcedir_template = getParams(labels_dict,"MRIQC_OUTPUT_DIR")

    try:
        # Link mriqc subject folder
        for part_value in subject_list:
            part_vals = part_value.split("^")
            project = part_vals[0]
            subject = part_vals[1]
            session = part_vals[2]
            labels_dict = updateParams(labels_dict,"PARTICIPANT_LABEL",subject)
            labels_dict = updateParams(labels_dict,"PARTICIPANT_XNAT_PROJECT",project)
            labels_dict = updateParams(labels_dict,"PARTICIPANT_SESSION",session)
            mriqc_sourcedir_sub = os.path.join(substitute_labels(mriqc_sourcedir_template,labels_dict),f"sub-{subject}")
            mriqc_targetdir_sub = os.path.join(mriqcgroup_outdir,f"sub-{subject}")
            # create symlink at subject level, this should create for first session as we sorted list
            if not os.path.exists(mriqc_targetdir_sub) and os.path.exists(mriqc_sourcedir_sub):
                os.symlink(mriqc_sourcedir_sub,mriqc_targetdir_sub)

            # A few cases have multiple sessions, we physically copy these the first time. this code will only really be run once
            # for each subject as subsequent runs will have the session physically in the original mriqc directory.
            # This is probably not an ideal approach but it keeps the individual mriqc runs for each session independent.
            mriqc_sourcedir_ses = os.path.join(substitute_labels(mriqc_sourcedir_template,labels_dict),f"sub-{subject}",f"ses-{session}")
            mriqc_targetdir_ses = os.path.join(mriqcgroup_outdir,f"sub-{subject}",f"ses-{session}")
            if not os.path.exists(mriqc_targetdir_ses) and os.path.exists(mriqc_sourcedir_ses):
                copytree(mriqc_sourcedir_ses,mriqc_targetdir_ses)
            
            
        # handle field maps - use dwirpe as standard if valid, otherwise use megre
        if not bids_dir:
            bids_dir = substitute_labels('<BIDS_DIR>', labels_dict)

        mriqc_dict={}
        mriqc_dict = updateParams(mriqc_dict,"-w","<CWD>/mriqcgroupwork")

        # Additional params
        MRIQC_OVERRIDE_PARAMS = getParams(labels_dict,"MRIQC_OVERRIDE_PARAMS")
        if MRIQC_OVERRIDE_PARAMS and isinstance(MRIQC_OVERRIDE_PARAMS,dict):
            add_labels(MRIQC_OVERRIDE_PARAMS,mriqc_dict)
    

        params = ""
        for mriqc_tag, mriqc_value in mriqc_dict.items():
            if "--" in mriqc_tag and "---" not in mriqc_tag:
                if mriqc_value == IS_PRESENT:
                    params=params + " " + mriqc_tag
                elif mriqc_value == IGNORE:
                    IFLOGGER.info(f"Parameter {mriqc_tag} is being skipped. This has been explicitly required in configuration.")
                else:
                    # we dont need = sign for mriqc just for basi;
                    params = params + " " + mriqc_tag + " " + mriqc_value

            elif "-" in mriqc_tag and "--" not in mriqc_tag:
                params = params + " " + mriqc_tag + " " + mriqc_value

            else:
                print(f"mriqc tag {mriqc_tag} not valid.")

        command=f"{command_base}"\
                " "+ bids_dir +\
                " "+ mriqcgroup_outdir +\
                " group"\
                " "+ params 

        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

        bold_rep = getGlob(os.path.join(cwd,'mriqcgroupout','*bold.tsv'))
        t1w_rep = getGlob(os.path.join(cwd,'mriqcgroupout','*T1w.tsv'))
        t2w_rep= getGlob(os.path.join(cwd,'mriqcgroupout','*T2w.tsv'))
        dwi_rep = getGlob(os.path.join(cwd,'mriqcgroupout','*dwi.tsv'))

        output_dir = cwd

        out_files=[]
        if bold_rep:
            out_files.insert(0,bold_rep)
        if t1w_rep:
            out_files.insert(0,t1w_rep)
        if t2w_rep:
            out_files.insert(0,t2w_rep)
        if dwi_rep:
            out_files.insert(0,dwi_rep)



        return {
            "output_dir":mriqcgroup_outdir,
            "out_files":out_files
        }
    except Exception as e:
        IFLOGGER.error(f"Exception created {e} - will try to unlink all mriqc subfolders at {mriqcgroup_outdir}")
        raise(e)
    finally:
        IFLOGGER.info(f"Attempting to unlink all subfolders at {mriqcgroup_outdir}")
        for part_value in subject_list:
            part_vals = part_value.split("^")
            subject = part_vals[1]
            mriqc_targetdir_sub = os.path.join(mriqcgroup_outdir,f"sub-{subject}")
            if os.path.islink(mriqc_targetdir_sub):
                os.unlink(mriqc_targetdir_sub)
        


class mriqcgroupInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class mriqcgroupOutputSpec(TraitedSpec):
    output_dir = traits.Directory(desc="MRIQC group output directory")
    out_files = traits.List(desc='list of files')
    
class mriqcgroup_pan(BaseInterface):
    input_spec = mriqcgroupInputSpec
    output_spec = mriqcgroupOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = mriqcgroup_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='mriqcgroup_node',bids_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(mriqcgroup_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import multiprocessing as mp
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging
from panpipelines.utils.report_functions import *

IFLOGGER=nlogging.getLogger('nipype.interface')

def reports_proc(labels_dict,analysis_level):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    html_file_dir = os.path.join(cwd,f"html_report")
    if not os.path.isdir(html_file_dir ):
        os.makedirs(html_file_dir,exist_ok=True)


    reports_source  = getParams(labels_dict,'REPORTS_SOURCE')
    if reports_source:
        reports_source = getGlob(reports_source)
    reports_pipeline_class = getParams(labels_dict,'REPORTS_PIPELINECLASS')
    reports_pipeline = getParams(labels_dict,'REPORTS_PIPELINE')
    if not reports_pipeline:
        reports_pipeline = "unknown-pipeline"


    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')
    if participant_label:
        html_name = f"sub-{participant_label}_ses-{session_label}_{reports_pipeline}.html"
    else:
        html_name = f"{reports_pipeline}.html"


    if reports_pipeline_class == "roiextract_panpipeline":
        html_file = newfile(outputdir=html_file_dir, assocfile = reports_source , suffix="htmlreport",extension="html")
        html_file = createRoiExtractReport(labels_dict,html_file, reports_source,analysis_level=analysis_level)
    elif reports_pipeline_class == "basil_panpipeline":
        html_file = newfile(outputdir=html_file_dir, assocfile = html_name, suffix="htmlreport",extension="html")
        html_file = createBasilReport(labels_dict,html_file, reports_source,analysis_level=analysis_level)


    out_files=[]
    out_files.insert(0,html_file)


    return {
        "html_file" : html_file,
        "out_files" : out_files
    }

class reportsInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    analysis_level = traits.String("participant",desc="analysis level", usedefault=True)

class reportsOutputSpec(TraitedSpec):
    html_file = File(desc='html file used for results')
    out_files = traits.List(desc='list of files')
    
class reports_pan(BaseInterface):
    input_spec = reportsInputSpec
    output_spec = reportsOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = reports_proc(
            self.inputs.labels_dict,
            self.inputs.analysis_level
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict, name='reports_node',LOGGER=IFLOGGER,analysis_level='participant'):
    # Create Node
    pan_node = Node(reports_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    pan_node.inputs.analysis_level = analysis_level
   
    return pan_node





from nipype import Workflow, MapNode, Node
from nipype.interfaces.io import DataSink

import panpipelines.nodes.antstransform as antstransform
import panpipelines.nodes.collate_csv_single as collate_csv
from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import glob

def create(name, wf_base_dir,labels_dict,createGraph=True,execution={}, LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)

    filecount_list=[]
    measures_list1=[]
    measures_template = getParams(labels_dict,"MEASURES_TEMPLATE")
    if isinstance(measures_template,list):
        for meas_template in measures_template:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            meas_files=glob.glob(evaluated_meas_template)
            measures_list1.extend(meas_files)
            filecount_list.append(len(meas_files))
    elif measures_template:
        measures_list1.extend(glob.glob(measures_template))


    collate_csv_node = collate_csv.create(labels_dict, csv_list1=measures_list1, LOGGER=LOGGER)


    sinker_dir = getParams(labels_dict,"SINKDIR")
    if sinker_dir:
        sinker = Node(DataSink(),name='collatecsv_sink')
        sinker_basedir = os.path.dirname(sinker_dir)
        sinker_folder = os.path.basename(sinker_dir)
        if not os.path.exists(sinker_basedir):
            os.makedirs(sinker_basedir,exist_ok=True)
        sinker.inputs.base_directory = sinker_basedir

        pan_workflow.connect( collate_csv_node,"roi_csv",sinker,f"{sinker_folder}")
        pan_workflow.connect( collate_csv_node,"roi_csv_metadata",sinker,f"{sinker_folder}.@metadata")

    else:
        pan_workflow.add_nodes([collate_csv_node])

    if createGraph:
         pan_workflow.write_graph(graph2use='flat')


    return pan_workflow



from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
from panpipelines.utils.upload_functions import *
import multiprocessing as mp
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def upload_proc(labels_dict,source_path,remote_path, ftpcredentials):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    pipeline = getParams(labels_dict,'PIPELINE')

    if ftpcredentials:
        with open(ftpcredentials,"r") as infile:
            ftpcred = json.load(infile)
            hostname=ftpcred["hostname"]
            username=ftpcred["username"]
            password=ftpcred["password"]
            if "port" in ftpcred.keys():
                port=int(ftpcred["password"])
            else:
                port = 22

    if source_path and remote_path and len(remote_path) == 1:
        IFLOGGER.info(f"Folders to upload {source_path} to {remote_path}")
        remote_path = [remote_path[0] for x in range(len(source_path))]
    elif not source_path:
        IFLOGGER.error(f"source_path not defined. cannot upload")
        raise Exception("upload.py : source_path not defined. Cannot upload")

    elif not remote_path:
        IFLOGGER.error(f"remote_path not defined. Cannot upload.")
        raise Exception("upload.py : remote_path not defined. Cannot upload")

    elif source_path and remote_path and not len(remote_path) == len(source_path):
        IFLOGGER.error(f"source path list {source_path} and remote path list {remote_path} do not match . Please ensure there is a 1:1 mapping or an N:1 mapping between source and remotes.")
        raise Exception("upload.py : length of source paths differs from remote paths and there are more than 1 remote paths.")

    retain_folder = isTrue( getParams(labels_dict,'RETAIN_FOLDER'))

    for source_path_dir,remote_path_dir in zip(source_path, remote_path):
        DEST_IS_DIR=False
        if os.path.isfile(source_path_dir):
            if remote_path_dir[-1]== "/":
                DEST_IS_DIR=True
                remote_path_dir = os.path.join(remote_path_dir,os.path.basename(source_path_dir))

            if retain_folder:
                remote_path_filename = os.path.basename(remote_path_dir)
                remote_path_dirname = os.path.dirname(remote_path_dir)
                source_path_folder = os.path.basename(os.path.dirname(source_path_dir))              
                remote_path_dir = os.path.join(remote_path_dirname,source_path_folder,remote_path_filename)
            

            ftp_upload(source_path_dir,remote_path_dir,hostname,username,password,port)
            remote_metadata_file = newfile(assocfile=os.path.dirname(remote_path_dir),suffix="upload-metadata",extension="json")
            local_metadata_file = newfile(outputdir=cwd,assocfile=remote_metadata_file,extension="json")
            
            metadata_init = {}
            history={}
            if DEST_IS_DIR:
                UPLOAD_SRC=source_path_dir
                UPLOAD_DEST=os.path.dirname(remote_path_dir )
            else:
                UPLOAD_SRC = source_path_dir
                UPLOAD_DEST = remote_path_dir 
            upload_metadata(local_metadata_file,remote_metadata_file,UPLOAD_SRC,UPLOAD_DEST,metadata=metadata_init,history=history,hostname=hostname,username=username,
            password=password, port=port)


        elif os.path.isdir(source_path_dir):
            if retain_folder:
                remote_path_dir = os.path.join(remote_path_dir,os.path.basename(source_path_dir))
            ftp_uploaddir_recursive(source_path_dir,remote_path_dir,hostname,username,password,port)

            remote_metadata_file = newfile(assocfile=os.path.dirname(remote_path_dir),suffix="upload-metadata",extension="json")
            local_metadata_file = newfile(outputdir=cwd,assocfile=remote_metadata_file,extension="json")
            
            metadata_init = {}
            history={}
            upload_metadata(local_metadata_file,remote_metadata_file,source_path_dir,remote_path_dir,metadata=metadata_init,history=history,hostname=hostname,username=username,
            password=password, port=port)

    out_files=[]
    out_files.append(local_metadata_file)

    return {
        "metadata_file" : local_metadata_file,
        "out_files" : out_files
    }

class uploadInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    source_path = traits.List([],desc="source path", usedefault=True)
    remote_path = traits.List([],desc="remote path", usedefault=True)
    ftpcredentials = traits.String("",desc="ftp credentials", usedefault=True)

class uploadOutputSpec(TraitedSpec):
    metadata_file = File(desc='metadata file for upload')
    out_files = traits.List(desc='list of files')

    
class upload_pan(BaseInterface):
    input_spec = uploadInputSpec
    output_spec = uploadOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = upload_proc(
            self.inputs.labels_dict,
            self.inputs.source_path,
            self.inputs.remote_path,
            self.inputs.ftpcredentials
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,source_path,remote_path, ftpcredentials, name='upload_node',LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(upload_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    pan_node.inputs.source_path = source_path
    pan_node.inputs.remote_path = remote_path
    pan_node.inputs.ftpcredentials = ftpcredentials
   
    return pan_node





from nipype import Workflow, Node
from nipype.interfaces.io import DataSink

import panpipelines.nodes.pancontainergroup as pancontainergroup
from panpipelines.utils.util_functions import *


def create(name, wf_base_dir,labels_dict,createGraph=True,execution={},LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)

    # Specify node inputs
    pancontainergroup_node = pancontainergroup.create(labels_dict,LOGGER=LOGGER)

    sinker_dir = getParams(labels_dict,"SINKDIR_GROUP")
    if sinker_dir:
        sinker = Node(DataSink(),name='collatecsvgroup_sink')
        sinker_basedir = os.path.dirname(sinker_dir)
        sinker_folder = os.path.basename(sinker_dir)
        if not os.path.exists(sinker_basedir):
            os.makedirs(sinker_basedir,exist_ok=True)
        sinker.inputs.base_directory = sinker_basedir
        pan_workflow.connect( pancontainergroup_node,"output_file",sinker,f"{sinker_folder}.@output_file")
        pan_workflow.connect( pancontainergroup_node,"metadata_file",sinker,f"{sinker_folder}.@metadata_file")
    else:
        pan_workflow.add_nodes([pancontainergroup_node])
    
    if createGraph:
         pan_workflow.write_graph(graph2use='flat')


    return pan_workflow



from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging
import json
import sys
import nibabel
from nibabel.affines import apply_affine 
import numpy as np
import numpy.linalg as npl
from scipy import ndimage 
from skimage import measure
import shutil
import panpipelines.scripts.pancontainer_panscript as pancontainer_script
import panpipelines.scripts.mne_make_surfaces as mne_make_surfaces
import mne
from panpipelines.nodes.antstransform import antstransform_proc
import itertools

# simulate strel('disk',5);
disc5 = np.zeros((9,9,9))
disc5[2:7,0:,0:]=1
disc5[0,2:7,2:7]=1
disc5[-1,2:7,2:7]=1
disc5[1,1:-1,1:-1]=1
disc5[7,1:-1,1:-1]=1

# simulate strel('disk',2);
disc2 = np.zeros((5,5,5))
disc2[2,2,2]=1
disc2[4,4,4]=1
disc2[2,0:,0:]=1
disc2[1,1:4,1:4]=1
disc2[3,1:4,1:4]=1
# simulate strel('disk',3)
disc3 = np.ones((5,5,5))

streldict={
    "disc5" : disc5,
    "disc3" : disc3,
    "disc2" : disc2
}


IFLOGGER=nlogging.getLogger('nipype.interface')

def returnCandidatesConservative(p0,p1,p2):
    arrayList=[]
    max_x = np.max(np.array([p0[0],p1[0],p2[0]]))
    max_y = np.max(np.array([p0[1],p1[1],p2[1]]))
    max_z = np.max(np.array([p0[2],p1[2],p2[2]]))
    min_x = np.min(np.array([p0[0],p1[0],p2[0]]))
    min_y = np.min(np.array([p0[1],p1[1],p2[1]]))
    min_z = np.min(np.array([p0[2],p1[2],p2[2]]))

    arrayList.extend([np.array(tup) for tup in itertools.product(range(min_x,max_x+1),range(min_y,max_y+1),range(min_z,max_z+1))])

    return arrayList

def returnCandidates(p0,p1,p2):
    arrayList=[]
    max_x = np.max(np.array([p0[0],p1[0],p2[0]]))
    max_y = np.max(np.array([p0[1],p1[1],p2[1]]))
    max_z = np.max(np.array([p0[2],p1[2],p2[2]]))
    min_x = np.min(np.array([p0[0],p1[0],p2[0]]))
    min_y = np.min(np.array([p0[1],p1[1],p2[1]]))
    min_z = np.min(np.array([p0[2],p1[2],p2[2]]))
    for x in range(min_x,max_x+1):
        for y in range(min_y, max_y + 1):
            for z in range(min_z,max_z + 1):
                p = np.array([x,y,z])
                if validCandidate(p0,p1,p2,p,tol=0.15):
                    arrayList.append(p)
    return arrayList

#https://math.stackexchange.com/questions/2582202/does-a-3d-point-lie-on-a-triangular-plane
def validCandidate(p0,p1,p2,p,tol):
    n = np.cross(p1-p0,p2-p0)/npl.norm(np.cross(p1-p0,p2-p0))
    d = np.dot(n,p-p1)
    if np.abs(d) < tol:
        return True
    else:
        return False

def create_circular_mask(h, w, center=None, radius=None):
    import numpy as np
    
    if center is None: # use the middle of the image
        center = (int(w/2), int(h/2))
    if radius is None: # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w-center[0], h-center[1])

    Y, X = np.ogrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0])**2 + (Y-center[1])**2)

    mask = dist_from_center <= radius
    return mask

def getDisc(dimx,dimy):
    disc2d = np.zeros((dimx,dimy))
    disc2d[create_circular_mask(dimx,dimy)]=1
    return disc2d

def get3Disc(dimN):
    npzero3d = np.zeros((dimN,dimN,dimN))
    disc2d = getDisc(dimN,dimN)
    for dimz in range(dimN):
        npzero3d[:,:,dimz] = np.multiply(np.multiply(np.ones((dimN,dimN)),disc2d[dimz,:]),disc2d[dimz,:].reshape(-1,1))

    return npzero3d

def within_range(vox,dims):
    if vox[0] > (dims[0] - 1) or vox[1] > (dims[1] - 1) or  vox[2] > (dims[2] - 1):
        return False
    
    return True

def derive_asl_artefact_v2(asl_acq,labels_dict,command_base,participant_label,participant_session,artefact_outputdir,PHASESHIFT=12,PHASEAXIS=1,ALLOWOVERLAP=False,CLOSE_DISC_STR="",DILATE_DISC_STR="disc-1",ERODE_DISC_STR="",T1_DILATE_DISC_STR="",COMBINED_ERODE_DISC_STR="",FILL_HOLES_STR="2",expand_ring=["1","1"],DO_T1_SHIFT=True, DO_CONSERVATIVE=True):


    transform_list =  getParams(labels_dict,'TRANSFORM_MAT')
    transform_ref =  getParams(labels_dict,'TRANSFORM_REF')
    PHASESHIFT_lookup = getParams(labels_dict,'PHASESHIFT')
    if PHASESHIFT_lookup:
        if isinstance(PHASESHIFT_lookup,dict):
            if asl_acq in PHASESHIFT_lookup.keys():
                PHASESHIFT=int(PHASESHIFT_lookup[asl_acq])
        else:
            PHASESHIFT = int(PHASESHIFT_lookup)
    if getParams(labels_dict,'PHASEAXIS'):
        PHASEAXIS = int(getParams(labels_dict,'PHASEAXIS'))
    if getParams(labels_dict,'CLOSE_DISC_STR'):
        CLOSE_DISC_STR = getParams(labels_dict,'CLOSE_DISC_STR')
    if getParams(labels_dict,'ERODE_DISC_STR'):
        ERODE_DISC_STR = getParams(labels_dict,'ERODE_DISC_STR')
    if getParams(labels_dict,'DILATE_DISC_STR'):
        DILATE_DISC_STR = getParams(labels_dict,'DILATE_DISC_STR')
    if getParams(labels_dict,'T1_DILATE_DISC_STR'):
        T1_DILATE_DISC_STR = getParams(labels_dict,'T1_DILATE_DISC_STR')
    if getParams(labels_dict,'FILL_HOLES_STR'):
        FILL_HOLES_STR = getParams(labels_dict,'FILL_HOLES_STR')
    if getParams(labels_dict,'COMBINED_ERODE_DISC_STR'):
        COMBINED_ERODE_DISC_STR = getParams(labels_dict,'COMBINED_ERODE_DISC_STR')
    if getParams(labels_dict,'DO_T1_SHIFT'):
        DO_T1_SHIFT= isTrue(getParams(labels_dict,'DO_T1_SHIFT'))
    if getParams(labels_dict,'ALLOWOVERLAP'):
        ALLOWOVERLAP = isTrue(getParams(labels_dict,'ALLOWOVERLAP'))
    if getParams(labels_dict,'DO_CONSERVATIVE'):
        DO_CONSERVATIVE = isTrue(getParams(labels_dict,'DO_CONSERVATIVE'))
    if getParams(labels_dict,'EXPANDRING'):
        expand_ring = getParams(labels_dict,'EXPANDRING')

    cwd = os.getcwd()
    output_dir=cwd

    work_dir=os.path.join(output_dir,"preproc_workdir")
    if not os.path.exists(work_dir):
        os.makedirs(work_dir,exist_ok=True)

    labels_dict = updateParams(labels_dict,"CWD",cwd)

    # set up dwi to process just the specific dwi session
    subject = f"sub-{participant_label}"
    if participant_session:
        session = f"ses-{participant_session}"
    else:
        session=""

    artefact_subdirname="asl_artefact"
    if getParams(labels_dict,'CHEMSHIFT_DIRNAME'):
        artefact_subdirname = substitute_labels(getParams(labels_dict,'CHEMSHIFT_DIRNAME'),labels_dict)

    if not artefact_outputdir:
        artefact_outputdir = os.path.join(output_dir,artefact_subdirname, subject,session)
    else:
        artefact_outputdir = os.path.join(artefact_outputdir,artefact_subdirname, subject,session)

    if not os.path.exists(artefact_outputdir):
        os.makedirs(artefact_outputdir, exist_ok=True)


    subjects_dir = getParams(labels_dict,'SUBJECTS_DIR')
    outer_skull=os.path.join(subjects_dir,subject,'bem','outer_skull.surf')

    params =  f" --subject {subject}" \
            f" --subjects_dir {subjects_dir}" 
    PREPROC_CONTAINER_TO_USE = getParams(labels_dict,"PREPROC_CONTAINER_TO_USE")
    if PREPROC_CONTAINER_TO_USE:
        panscript = pancontainer_script.pancontainer_panscript(labels_dict,params=params,command=f"python {mne_make_surfaces.__file__}",container_img=PREPROC_CONTAINER_TO_USE)
    else: 
        panscript = pancontainer_script.pancontainer_panscript(labels_dict,params=params,command=f"python {mne_make_surfaces.__file__}")
    panscript.run()
    os.chdir(output_dir)

    rr_mm_outer_skull,tris_outer_skull = mne.read_surface(outer_skull)

    orig_file=f"{subjects_dir}/{subject}/mri/T1.mgz"
    origimg = nibabel.load(orig_file)
    Torig = origimg.header.get_vox2ras_tkr()
    origimg_dtype = origimg.header.get_data_dtype()
    origimg_shape = origimg.shape
    outer_skull_vol = np.zeros((origimg_shape),dtype=origimg_dtype)
    inv_Torig = npl.inv(Torig)

    for rr in rr_mm_outer_skull:
        vox = apply_affine(inv_Torig, rr)
        vox_ind = tuple(np.round(vox).astype(int))
        if within_range(vox_ind,origimg_shape):
            outer_skull_vol[vox_ind] = 255
        else:
            IFLOGGER.warn(f"{vox} outside range of {origimg_shape} for surface {rr}")

    for tri in tris_outer_skull:
        p0 = np.round(apply_affine(inv_Torig,rr_mm_outer_skull[tri[0]])).astype(int)
        p1 = np.round(apply_affine(inv_Torig,rr_mm_outer_skull[tri[1]])).astype(int)
        p2 = np.round(apply_affine(inv_Torig,rr_mm_outer_skull[tri[2]])).astype(int)
        if DO_CONSERVATIVE:
            valid_vox = returnCandidatesConservative(p0,p1,p2)
        else:
            valid_vox = returnCandidates(p0,p1,p2)
        for vox in valid_vox:
            vox_ind = tuple(vox.astype(int))
            if within_range(vox_ind,origimg_shape):
                outer_skull_vol[vox_ind] = 255
            else:
                IFLOGGER.warn(f"{vox} outside range of {origimg_shape} for surface {rr}")
    outer_skull_img = nibabel.Nifti1Image(outer_skull_vol,origimg.affine,origimg.header)
    outer_skull_img_file=os.path.join(work_dir,f"{subject}_{session}_outer_skull.nii.gz")
    nibabel.save(outer_skull_img,outer_skull_img_file) 

    if T1_DILATE_DISC_STR:
        T1_DILATE_DISC=int(T1_DILATE_DISC_STR.split("-")[1])
        STREL_TYPE=T1_DILATE_DISC_STR.split("-")[0]

        if T1_DILATE_DISC > 0:
            if STREL_TYPE.upper() == "DISC":
                disc = get3Disc(T1_DILATE_DISC)
            else:
                disc = np.ones((T1_DILATE_DISC,T1_DILATE_DISC,T1_DILATE_DISC))

            outer_skull_vol_dilation = ndimage.binary_dilation(outer_skull_vol, structure=disc).astype(np.int16)
            outer_skull_vol_dilation[outer_skull_vol_dilation > 0] = 255
            outer_skull_vol_dilation_file=newfile(work_dir, outer_skull_img_file,suffix=f"t1dilation-{T1_DILATE_DISC}")
            outer_skull_vol_dilation_img = nibabel.Nifti1Image(outer_skull_vol_dilation,outer_skull_img.affine,outer_skull_img.header)
            nibabel.save(outer_skull_vol_dilation_img,outer_skull_vol_dilation_file)
            outer_skull_img_file = outer_skull_vol_dilation_file
            outer_skull_vol = outer_skull_vol_dilation

    t1_out_dims = outer_skull_vol.shape
    if DO_T1_SHIFT:
        asl_ref = getParams(labels_dict,'TRANSFORM_REF')
        asl_img = nibabel.load(getGlob(asl_ref))
        t1_ori = get_orientation_from_file(outer_skull_img_file,"image")[0]
        T1_PHASEAXIS  = t1_ori.index("A")

        asl_dim = asl_img.header["pixdim"][PHASEAXIS]
        t1_dim = outer_skull_img.header["pixdim"][T1_PHASEAXIS]

        T1_PHASESHIFTFACTOR=int(np.round(asl_dim/t1_dim))
        T1_PHASESHIFT = PHASESHIFT * T1_PHASESHIFTFACTOR

        if T1_PHASESHIFT < 0:
            T1_PHASESHIFT_CALC = t1_out_dims[T1_PHASEAXIS] + T1_PHASESHIFT
            T1_OVSTART=T1_PHASESHIFT_CALC
            T1_OVEND=-1
        else:
            T1_PHASESHIFT_CALC = T1_PHASESHIFT
            T1_OVSTART=0
            T1_OVEND=T1_PHASESHIFT

        outer_skull_t1_shift=newfile(work_dir,outer_skull_img_file,suffix=f"t1shift_{T1_PHASESHIFT}")
        T1_data_shifted = np.roll(outer_skull_vol,T1_PHASESHIFT_CALC,T1_PHASEAXIS)
        if not ALLOWOVERLAP:
            if T1_PHASEAXIS==0:
                T1_data_shifted[T1_OVSTART:T1_OVEND,:,:]=0
            elif T1_PHASEAXIS==1:
                T1_data_shifted[:,T1_OVSTART:T1_OVEND,:]=0
            elif T1_PHASEAXIS==2:
                T1_data_shifted[:,:,T1_OVSTART:T1_OVEND]=0
            else:
                print(f"phase axis {PHASEAXIS} not valid. Allowing overlap of mask")     
        t1_shifted_img = nibabel.Nifti1Image(T1_data_shifted,outer_skull_img.affine, outer_skull_img.header)

        nibabel.save(t1_shifted_img,outer_skull_t1_shift)
        outer_skull_img_file = outer_skull_t1_shift
        outer_skull_vol = T1_data_shifted

    outer_skull_aslspace_file = None
    if transform_list and transform_ref:
        results = antstransform_proc(labels_dict, outer_skull_img_file,transform_list, transform_ref)
        outer_skull_aslspace_file = results['out_file']     

    outer_skull_aslspace_bin=newfile(work_dir,outer_skull_aslspace_file,suffix="bin")
    command = f"{command_base} fslmaths {outer_skull_aslspace_file} -abs -bin {outer_skull_aslspace_bin}"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    outskull_aslspace_img=nibabel.load(outer_skull_aslspace_bin)
    outskull_aslspace_data = outskull_aslspace_img.get_fdata()
    out_dims = outskull_aslspace_data.shape
    outskull_aslspace_file  = outer_skull_aslspace_bin

    if DILATE_DISC_STR:
        DILATE_DISC=int(DILATE_DISC_STR.split("-")[1])
        STREL_TYPE=DILATE_DISC_STR.split("-")[0]
        if DILATE_DISC > 0:
            if STREL_TYPE.upper() == "DISC":
                disc = get3Disc(DILATE_DISC)
            else:
                disc = np.ones((DILATE_DISC,DILATE_DISC,DILATE_DISC))    
            outskull_aslspace_data_dilation = ndimage.binary_dilation(outskull_aslspace_data, structure=disc).astype(np.int16)
            outskull_aslspace_data_dilation_file=newfile(work_dir,outskull_aslspace_file,suffix=f"dilation-{DILATE_DISC}")
            outskull_aslspace_data_dilation_img = nibabel.Nifti1Image(outskull_aslspace_data_dilation,outskull_aslspace_img.affine,outskull_aslspace_img.header)
            nibabel.save(outskull_aslspace_data_dilation_img,outskull_aslspace_data_dilation_file)
            outskull_aslspace_file = outskull_aslspace_data_dilation_file
            outskull_aslspace_data = outskull_aslspace_data_dilation

    if FILL_HOLES_STR:
        FILL_HOLES=int(FILL_HOLES_STR)
        # Assume LAS or RAS!
        ori = get_orientation_from_file(outer_skull_aslspace_file,"image")[0]
        if ori == "LAS" or ori == "RAS":
            holes_filled=0
            for outdim in range(out_dims[1]):
                if holes_filled == FILL_HOLES:
                    break
                if np.sum(outskull_aslspace_data[:,outdim,:]> 0) > 3:
                    outskull_aslspace_data[:,outdim,:] = ndimage.binary_fill_holes(outskull_aslspace_data[:,outdim,:])
                    holes_filled = holes_filled + 1
  
        outskull_aslspace_data_filled_file=newfile(work_dir,outskull_aslspace_file,suffix=f"filled-{FILL_HOLES}")
        outskull_aslspace_data_filled_img = nibabel.Nifti1Image(outskull_aslspace_data,outskull_aslspace_img.affine,outskull_aslspace_img.header)
        nibabel.save(outskull_aslspace_data_filled_img,outskull_aslspace_data_filled_file)
        outskull_aslspace_file = outskull_aslspace_data_filled_file

    if ERODE_DISC_STR:
        ERODE_DISC=int(ERODE_DISC_STR.split("-")[1])
        STREL_TYPE=ERODE_DISC_STR.split("-")[0]

        if ERODE_DISC > 0:
            if STREL_TYPE.upper() == "DISC":
                disc = get3Disc(ERODE_DISC)
            else:
                disc = np.ones((ERODE_DISC,ERODE_DISC,ERODE_DISC))

            outskull_aslspace_data_erosion = ndimage.binary_erosion(outskull_aslspace_data, structure=disc).astype(np.int16)
            outskull_aslspace_data_erosion_file=newfile(work_dir,outskull_aslspace_file,suffix=f"erosion-{ERODE_DISC}")
            outskull_aslspace_data_erosion_img = nibabel.Nifti1Image(outskull_aslspace_data_erosion,outskull_aslspace_img.affine,outskull_aslspace_img.header)
            nibabel.save(outskull_aslspace_data_erosion_img,outskull_aslspace_data_erosion_file)
            outskull_aslspace_file = outskull_aslspace_data_erosion_file
            outskull_aslspace_data = outskull_aslspace_data_erosion
    
    if CLOSE_DISC_STR:
        CLOSE_DISC=int(CLOSE_DISC_STR.split("-")[1])
        STREL_TYPE=CLOSE_DISC_STR.split("-")[0]

        if CLOSE_DISC > 0:
            if STREL_TYPE.upper() == "DISC":
                disc = get3Disc(CLOSE_DISC)
            else:
                disc = np.ones((CLOSE_DISC,CLOSE_DISC,CLOSE_DISC))

            outskull_aslspace_data_close = ndimage.binary_closing(outskull_aslspace_data, structure=disc).astype(np.int16)
            outskull_aslspace_data_close_file=newfile(work_dir,outskull_aslspace_file,suffix=f"close-{CLOSE_DISC}")
            outskull_aslspace_data_close_img = nibabel.Nifti1Image(outskull_aslspace_data_close,outskull_aslspace_img.affine,outskull_aslspace_img.header)
            nibabel.save(outskull_aslspace_data_close_img,outskull_aslspace_data_close_file)
            outskull_aslspace_file = outskull_aslspace_data_close_file
            outskull_aslspace_data = outskull_aslspace_data_close

    if len(expand_ring) == 1:
        prering=int(expand_ring[0])
        phase_range=range(PHASESHIFT - prering,PHASESHIFT + 1)
        if DO_T1_SHIFT:
            phase_range=range(-prering,1)
    elif len(expand_ring)> 1:
        prering=int(expand_ring[0])
        postring=int(expand_ring[1])
        phase_range=range(PHASESHIFT - prering,PHASESHIFT + 1 + postring)
        if DO_T1_SHIFT:
            phase_range=range(-prering,1 + postring)
    else:
        phase_range=range(PHASESHIFT,PHASESHIFT + 1)
        if DO_T1_SHIFT:
            phase_range=range(0,1)

    outer_skull_shifted_images = []

    COMBINED=False
    if len(phase_range) > 1:
        COMBINED=True 

    for phaseshift in phase_range:
        if phaseshift < 0:
            PHASESHIFT_CALC = out_dims[PHASEAXIS] + phaseshift
            OVSTART=PHASESHIFT_CALC
            OVEND=-1
        else:
            PHASESHIFT_CALC = phaseshift
            OVSTART=0
            OVEND=phaseshift

        outer_skull_shift=newfile(work_dir,outskull_aslspace_file,suffix=f"shift_{phaseshift}")
        data_shifted = np.roll(outskull_aslspace_data,PHASESHIFT_CALC,PHASEAXIS)
        if not ALLOWOVERLAP:
            if PHASEAXIS==0:
                data_shifted[OVSTART:OVEND,:,:]=0
            elif PHASEAXIS==1:
                data_shifted[:,OVSTART:OVEND,:]=0
            elif PHASEAXIS==2:
                data_shifted[:,:,OVSTART:OVEND]=0
            else:
                print(f"phase axis {PHASEAXIS} not valid. Allowing overlap of mask")     
        shifted_img = nibabel.Nifti1Image(data_shifted,outskull_aslspace_img.affine,outskull_aslspace_img.header)
        nibabel.save(shifted_img,outer_skull_shift)
        outer_skull_shifted_images.append(outer_skull_shift)

    if len(outer_skull_shifted_images) > 1:
        command = f"{command_base} fslmaths {outer_skull_shifted_images[0]} "
        for shifted_image in outer_skull_shifted_images[1:]:
            command = command + f"-add {shifted_image} "

        outer_skull_shift_combined=newfile(work_dir,outer_skull_shift,suffix=f"combined")
        command = command + f"-bin {outer_skull_shift_combined} "
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    else:
        outer_skull_shift_combined =  outer_skull_shift

    if COMBINED_ERODE_DISC_STR and COMBINED:
        COMBINED_ERODE_DISC=int(COMBINED_ERODE_DISC_STR.split("-")[1])
        STREL_TYPE=COMBINED_ERODE_DISC_STR.split("-")[0]

        if CLOSE_DISC > 0:
            if STREL_TYPE.upper() == "DISC":
                disc = get3Disc(COMBINED_ERODE_DISC)
            else:
                disc = np.ones((COMBINED_ERODE_DISC,COMBINED_ERODE_DISC,COMBINED_ERODE_DISC))

            outer_skull_shift_combined_img=nibabel.load(outer_skull_shift_combined)
            outer_skull_shift_combined_data = outer_skull_shift_combined_img.get_fdata()
            
             
            outer_skull_shift_combined_erosion_data = ndimage.binary_erosion(outer_skull_shift_combined_data, structure=disc).astype(np.int16)
            outer_skull_shift_combined_erosion_file=newfile(work_dir,outer_skull_shift_combined,suffix=f"comb-erosion-{COMBINED_ERODE_DISC}")
            outer_skull_shift_combined_erosion_img = nibabel.Nifti1Image(outer_skull_shift_combined_erosion_data,outer_skull_shift_combined_img.affine,outer_skull_shift_combined_img.header)
            nibabel.save(outer_skull_shift_combined_erosion_img,outer_skull_shift_combined_erosion_file)
            outer_skull_shift_combined = outer_skull_shift_combined_erosion_file

    outer_skull_shift_final = newfile(artefact_outputdir,assocfile=f"{subject}_{session}_asl_chemical_shift_artefact.nii.gz")
    shutil.copyfile(outer_skull_shift_combined,outer_skull_shift_final)  


def derive_asl_artefact_v1(asl_acq,labels_dict,command_base,aslfile,aslfile_brain,aslfile_brain_mask,workdir,outputdir,RINGTHRESH=0.15,STRELCLOSE=disc3,PHASESHIFT=12,PHASEAXIS=1,FRAC_INT_THRESH="0.7000000000000002",VERT_GRAD="0"):

    FRAC_INT_THRESH_lookup= getParams(labels_dict,'FRAC_INT_THRESH')
    if FRAC_INT_THRESH_lookup:
        FRAC_INT_THRESH = FRAC_INT_THRESH_lookup

    VERT_GRAD_lookup=getParams(labels_dict,'VERT_GRAD')
    if VERT_GRAD_lookup:
        VERT_GRAD = VERT_GRAD_lookup

    RINGTHRESH_lookup=getParams(labels_dict,'RINGTHRESH')
    if RINGTHRESH_lookup:
        RINGTHRESH = np.float64(RINGTHRESH_lookup)

    STRELCLOSE_lookup=getParams(labels_dict,'STRELCLOSE')
    if STRELCLOSE_lookup and STRELCLOSE_lookup in streldict.keys():
        STRELCLOSE = streldict[STRELCLOSE_lookup]

    PHASESHIFT_lookup = getParams(labels_dict,'PHASESHIFT')
    if PHASESHIFT_lookup:
        if isinstance(PHASESHIFT_lookup,dict):
            if asl_acq in PHASESHIFT_lookup.keys():
                PHASESHIFT=int(PHASESHIFT_lookup[asl_acq])
        else:
            PHASESHIFT = int(PHASESHIFT_lookup)
    if getParams(labels_dict,'PHASEAXIS'):
        PHASEAXIS = int(getParams(labels_dict,'PHASEAXIS'))

    params=f" {aslfile}" \
        f" {aslfile_brain}" \
        " -R" \
        f" -f {FRAC_INT_THRESH}" \
        f" -g {VERT_GRAD}" \
        " -m -s"

    command=f"{command_base} bet"\
        " "+params
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    data1 = nibabel.load(aslfile).get_fdata()
    data1 = data1[:,:,:,0]
    data2img = nibabel.load(aslfile_brain)
    data2 = data2img.get_fdata()
    data3img = nibabel.load(aslfile_brain_mask)
    data3 = data3img.get_fdata()

    diff12 = np.abs(data1-data2)
    diff12_img = nibabel.Nifti1Image(diff12,data2img.affine,data2img.header)
    ring = newfile(workdir,assocfile=aslfile,suffix="ring")
    nibabel.save(diff12_img,ring)

    diff12bw = diff12.copy()
    diff12bw_mask = diff12[:,:,:] > RINGTHRESH*np.max(diff12)
    diff12bw[diff12bw_mask]=1
    diff12bw[~diff12bw_mask]=0
    diff12bw = np.int16(diff12bw)
    diff12bw_shifted = np.roll(diff12bw,PHASESHIFT,PHASEAXIS)
    diff12bw_shifted_img = nibabel.Nifti1Image(diff12bw_shifted,data3img.affine,data3img.header)
    ring_shifted = newfile(workdir,assocfile=aslfile,suffix="ring_shifted")
    nibabel.save(diff12bw_shifted_img,ring_shifted)

    overlap_voxels = diff12bw_shifted + data3
    overlap_voxels_mask = overlap_voxels[:,:,:] == 2
    overlap_voxels[overlap_voxels_mask] = 1
    overlap_voxels[~overlap_voxels_mask] = 0
    overlap_voxels = np.int16(overlap_voxels)
    overlap_img = nibabel.Nifti1Image(overlap_voxels,data3img.affine,data3img.header)
    overlap = newfile(workdir,assocfile=aslfile,suffix="artefact_premorph")
    nibabel.save(overlap_img,overlap)

    overlap_voxels_morpho = ndimage.binary_closing(overlap_voxels, structure=STRELCLOSE).astype(np.int16)

    labels = measure.label(overlap_voxels_morpho)
    props = measure.regionprops(labels)
    largest_prop = max(props, key=lambda prop: prop.area)
    overlap_voxels_morpho_clean=np.zeros(overlap_voxels_morpho.shape)
    overlap_voxels_morpho_clean[tuple(largest_prop.coords.T)] = 1
    overlap_voxels_morpho_clean = np.int16(overlap_voxels_morpho_clean)
    overlap_clean_img = nibabel.Nifti1Image(overlap_voxels_morpho_clean,data3img.affine,data3img.header)
    overlap_clean = newfile(workdir,assocfile=aslfile,suffix="artefact")
    nibabel.save(overlap_clean_img,overlap_clean)

    final_clean = newfile(outputdir,assocfile=overlap_clean)
    shutil.copyfile(overlap_clean,final_clean)


def preproc_proc(labels_dict,bids_dir=""):

    cwd=os.getcwd()
    output_dir = cwd

    work_dir=os.path.join(output_dir,"preproc_workdir")
    if not os.path.exists(work_dir):
        os.makedirs(work_dir,exist_ok=True)

    labels_dict = updateParams(labels_dict,"CWD",cwd)

    command_base, container = getContainer(labels_dict,nodename="preproc", SPECIFIC="FSL_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the fsl version:")
    command = f"{command_base} fslversion"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    # set up dwi to process just the specific dwi session
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    subject = f"sub-{participant_label}"
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')
    session = f"ses-{participant_session}"
    preproc_version = getParams(labels_dict,'PREPROC_VERSION')
    if preproc_version:
        VERSION_TO_RUN = preproc_version
    else:
        VERSION_TO_RUN = "2"

    # calculate asl artefact for acq-prod
    layout = BIDSLayout(bids_dir)
    asl=layout.get(subject=participant_label,session=participant_session,suffix='asl', extension='nii.gz')

    if len(asl) > 0:
        asl_bidsfile=asl[0]
        asl_entities = asl_bidsfile.get_entities()
        if "acquisition" in asl_entities.keys():
            asl_acq = "acq-" + asl_entities["acquisition"]
        else:
            asl_acq = get_bidstag("acq",asl_bidsfile.filename)

        if not asl_acq:
            asl_acq = "default"

        artefact_outputdir = getParams(labels_dict,'DERIVATIVES_DIR')
        if not artefact_outputdir:
            artefact_outputdir = output_dir


        if VERSION_TO_RUN == "2":
            derive_asl_artefact_v2(asl_acq,labels_dict,command_base, participant_label,participant_session,artefact_outputdir)
        else:
            m0_entities = asl_entities.copy()
            m0_entities["suffix"]="m0scan"
            m0  = layout.get(return_type='file', invalid_filters='allow', **m0_entities)
            if len(m0) > 0:
                m0_file=m0[0]
                m0_file_brain = newfile(work_dir,assocfile=m0_file,suffix="brain")
                m0_file_brain_mask = newfile(work_dir,assocfile=m0_file_brain,suffix="mask")
                derive_asl_artefact_v1(asl_acq,labels_dict, command_base, m0_file,m0_file_brain,m0_file_brain_mask,work_dir,artefact_outputdir)

    return {
        "output_dir":output_dir
    }



class preprocInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class preprocOutputSpec(TraitedSpec):
    output_dir = traits.String(desc="QSIPREP output directory")
    
    
class preproc_pan(BaseInterface):
    input_spec = preprocInputSpec
    output_spec = preprocOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = preproc_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='preproc_node',bids_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(preproc_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
from panpipelines.utils import transformer as tran
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging
from shutil import copy

IFLOGGER=nlogging.getLogger('nipype.interface')

def xtract_proc(labels_dict,input_dir):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')


    output_dir = os.path.join(cwd,'xtract_results')
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir,exist_ok=True)

    xtractout_dir = os.path.join(output_dir,'xtract_out')
    if not os.path.isdir(xtractout_dir):
        os.makedirs(xtractout_dir,exist_ok=True)

    dmri_dir = os.path.join(output_dir,'qsidmri')
    if not os.path.isdir(dmri_dir):
        os.makedirs(dmri_dir,exist_ok=True)

    transforms_dir = os.path.join(output_dir,'transforms')
    if not os.path.isdir(transforms_dir):
        os.makedirs(transforms_dir,exist_ok=True)

    work_dir = os.path.join(cwd,'xtract_workdir')
    if not os.path.isdir(work_dir):
        os.makedirs(work_dir,exist_ok=True)

    input_dir=substitute_labels(input_dir,labels_dict)
    dwi = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(session_label),'dwi','*_desc-preproc_dwi.nii.gz'))
    dwiref = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(session_label),'dwi','*_dwiref.nii.gz'))
    mask = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(session_label),'dwi','*_desc-brain_mask.nii.gz'))
    grad = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(session_label),'dwi','*_desc-preproc_dwi.b'))

    mr_command_base, mr_container = getContainer(labels_dict,nodename="xtract", SPECIFIC="MRTRIX_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the mrtrix version:")
    command = f"{mr_command_base} mrinfo --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    if mr_container:
        IFLOGGER.info("\nChecking the container version:")
        command = f"{mr_container} --version"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    IFLOGGER.info("Preparing data from qsiprep for bedpost run")
    dwitrix = newfile(outputdir=work_dir, assocfile=dwi, extension=".mif")
    command=f"{mr_command_base} mrconvert {dwi} {dwitrix} -grad {grad}"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    bvec=os.path.join(dmri_dir,"bvecs")
    bval=os.path.join(dmri_dir,"bvals")
    command=f"{mr_command_base} dwigradcheck {dwitrix} -export_grad_fsl {bvec} {bval}"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    dwipreproc = os.path.join(dmri_dir,"data.nii.gz")
    command=f"{mr_command_base} mrconvert {dwitrix} {dwipreproc}"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    nodif_brain_mask = os.path.join(dmri_dir,"nodif_brain_mask.nii.gz")
    copy(mask,nodif_brain_mask)

    command_base, container = getContainer(labels_dict,nodename="xtract", SPECIFIC="XTRACT_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the fsl version:")
    command = f"{command_base} fslversion"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    if container:
        IFLOGGER.info("\nChecking the container version:")
        command = f"{container} --version"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    IFLOGGER.info("Run bedpostx_gpu")
    command=f"{command_base} bedpostx_gpu {dmri_dir} -n 3 -model 2"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    IFLOGGER.info("Calculate transforms to MNI152 space")

    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    if not TEMPLATEFLOW_HOME:
        TEMPLATEFLOW_HOME=os.path.abspath("TemplateFlow")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

    mnilin6_res1_ras = str(tran.get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin6Asym",resolution=1,suffix="T1w",extension=[".nii.gz"]))
    mnilin6_res1_las= newfile(outputdir=work_dir,assocfile=mnilin6_res1_ras,suffix="ori-LAS")
    tran.reorient(mnilin6_res1_ras,"LAS",mnilin6_res1_las)

    mni2009_res1_ras = str(tran.get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin2009cAsym",resolution=1,suffix="T1w",extension=[".nii.gz"]))
    mni2009_res1_las=newfile(outputdir=work_dir,assocfile=mni2009_res1_ras,suffix="ori-LAS")
    tran.reorient(mni2009_res1_ras,"LAS",mni2009_res1_las)

    mni2009_mnilin6_warp = str(tran.get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin6Asym",suffix="xfm",extension=[".h5"]))
    t1acpc_mni2009_warp=str(getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'anat','*_from-T1w_to-MNI*')))

    os.chdir(work_dir)
    t1 = tran.disassembleTransforms(t1acpc_mni2009_warp,"t1",command_base)
    t1.sort()
    t2 = tran.disassembleTransforms(mni2009_mnilin6_warp,"t2",command_base)
    t2.sort()
    os.chdir(cwd)

    fulltransform=[]
    fulltransform.append(t1[0])
    fulltransform.append(t1[1])
    fulltransform.append(t2[0])
    fulltransform.append(t2[1])

    dwiacpc_mni152_warp = newfile(outputdir=work_dir,assocfile="dwiacpc_mnilin6_warp",extension="nii.gz")
    tran.apply_transform_ants(dwiref,mnilin6_res1_las,dwiacpc_mni152_warp,fulltransform,command_base,composite=True)

    # use mr_command_base as wb_command_base has issue with --nv gpu access
    diff2std_warp = newfile(outputdir=transforms_dir,assocfile="diff2std_warp",extension="nii.gz")
    tran.convertWarp_toFNIRT(dwiacpc_mni152_warp, diff2std_warp, dwiref, mr_command_base)

    std2diff_warp = newfile(outputdir=transforms_dir,assocfile="std2diff_warp",extension="nii.gz")
    tran.invertWarpfield_FNIRT(dwiref,diff2std_warp,std2diff_warp ,command_base)

    IFLOGGER.info("Run xtract")
    command=f"{command_base} xtract -bpx {dmri_dir}.bedpostX -out {xtractout_dir} -species HUMAN -stdwarp {std2diff_warp} {diff2std_warp} -gpu "
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    THRESH = getParams(labels_dict,"XTRACT_THRESH")
    if not THRESH:
        THRESH="0.001"

    tracts = os.path.join(xtractout_dir,"tracts")
    for tractdir in os.listdir(tracts):
        tract = os.path.join(tracts,tractdir,"densityNorm.nii.gz")
        tract_acpc_dir = os.path.join(os.path.dirname(tracts),"tracts_acpc")
        os.makedirs(tract_acpc_dir,exist_ok=True)
        tract_acpc = newfile(outputdir=tract_acpc_dir,assocfile=tract,prefix=tractdir,suffix="space-acpc")
        tran.applyWarp_fnirt(tract,dwiref,tract_acpc,std2diff_warp,command_base,interp="nn")

        tract_acpc_thresh = newfile(outputdir=tract_acpc_dir,assocfile=tract_acpc,suffix=f"desc-thresh{THRESH}") 
        command=f"{command_base} fslmaths {tract} -thr {THRESH} -bin {tract_acpc_thresh}"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

        tran.applyWarp_fnirt(tract_acpc_thresh,dwiref,tract_acpc_thresh,std2diff_warp,command_base,interp="nn",relative=True)

    out_files=[]

    return {
        "output_dir":output_dir,
        "out_files":out_files
    }



class xtractInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_dir = traits.String("",desc="QSIPREP Output Directory", usedefault=True)

class xtractOutputSpec(TraitedSpec):
    output_dir = traits.String(desc="XTRACT output directory")
    out_files = traits.List(desc='list of files')
    
class xtract_pan(BaseInterface):
    input_spec = xtractInputSpec
    output_spec = xtractOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = xtract_proc(
            self.inputs.labels_dict,
            self.inputs.input_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="xtract_node",input_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(xtract_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if input_dir is None:
        input_dir = ""
        
    pan_node.inputs.input_dir =  input_dir

    return pan_node


